# vim: set tabstop=4 shiftwidth=4 fdm=marker fmr=<<<,>>>
package GPST12N;
use Carp;
use strict;
use warnings;
use List::Util  qw/min max reduce/;
use Data::Dumper;
use Smart::Comments;

my @cachedTriangles;

sub getTransformedWaypoints{ # <<<2
	my $supportpoints = shift;
	my $waypoints     = shift;
	my ($window_zoom,$window_offx,$window_offy,$large_width) = @_;

	return getScaledWaypoints($waypoints,$window_zoom,$window_offx,$window_offy,$large_width)
		unless(scalar keys %$supportpoints > 4);

	my %ret;

	### triangulate GPS-space using 'triangle'...
	my %gpssupportpoints = map {$_ => $waypoints->{$_}} keys %$supportpoints;

	determineSurroundingSupportPoints(\%gpssupportpoints,$supportpoints);
	map{$waypoints->{$_}=$gpssupportpoints{$_}}grep{/^tmp/}keys %gpssupportpoints;

	@cachedTriangles     = TriangleWrap::getTriangles(\%gpssupportpoints);

	if (0){
		open T, ">joerg/triangles.dat" or die $!;
		print T join(' ', (map{@$_}$_)),"\n" foreach(@cachedTriangles);
		close T;
		open T, ">joerg/support.dat" or die $!;
		print T join(' ', $_, @{$gpssupportpoints{$_}}, @{$supportpoints->{$_}}), "\n" foreach (keys %gpssupportpoints);
		close T;
	}

	
	### interpolating waypoints...;
	TRANSFORM_WP: 
	foreach my $p (keys(%$waypoints), grep{/^tmp/}keys(%gpssupportpoints)){
		if (exists $supportpoints->{$p}){
			my($x,$y) = @{$supportpoints->{$p}};
			$x = int($x/$window_zoom - $window_offx/$window_zoom + 0.5);
			$y = int($y/$window_zoom - $window_offy/$window_zoom + 0.5);
			$ret{$p} = [$x,$y];
			next;
		}
		@_ = @{$waypoints->{$p}};
		@_ = getTransformed(\%gpssupportpoints,$supportpoints,\@cachedTriangles,@_);
		if($_[0] == 0 and $_[1] == 0) {
			warn "Cannot get coordinates for $p:  @_";
		}
		$ret{$p} = [
			int($_[0]/$window_zoom-$window_offx/$window_zoom+0.5),
			int($_[1]/$window_zoom-$window_offy/$window_zoom+0.5)];
	}
	return %ret;
}
sub getTransformed{
	my $origSupportPoints = shift;
	my $destSupportPoints = shift;
	my $triangles         = shift;
	my @coor = @_;
	foreach my $t (@$triangles){
		die "$t is not a triangle" unless ref $t;
		my @tmp = @$origSupportPoints{@$t};
		@_ = (@coor, map{@$_}(@tmp));
		#print "checking whether ", join ";", @_, "\n";
		if(Math::NumberCruncher::InTriangle(@_)){
			#print "@coor is inside ", join(", ",@$t), "; interpolating...\n";
			my ($l0,$l1,$l2) = BaryCentric::getInterpolationFactors(@_);
			my $x = $l0 * $destSupportPoints->{$t->[0]}->[0]
					+$l1 * $destSupportPoints->{$t->[1]}->[0]
					+$l2 * $destSupportPoints->{$t->[2]}->[0];
			my $y = $l0 * $destSupportPoints->{$t->[0]}->[1]
					+$l1 * $destSupportPoints->{$t->[1]}->[1]
					+$l2 * $destSupportPoints->{$t->[2]}->[1];
			#print "set @coor to $x, $y\n";
			return ($x, $y);
		}
	}
	warn "Cannot find triangle for coordinate @coor";
	return (0,0);
}
sub img2gps{
	my $supportpoints = shift;
	my $waypoints     = shift;
	my @coor = @_;
	unless(@cachedTriangles){
		die "you need to triangulate the space before calling this function";
	}
	return getTransformed($supportpoints,$waypoints,\@cachedTriangles,@coor);
}
sub getScaledWaypoints{ # <<<2
	my $waypoints = shift;
	my ($window_zoom,$window_offx,$window_offy,$large_width) = @_;
	my %ret;
	# preprocess the waypoints for displaying
	foreach my $k (keys %$waypoints){
	  my ($lat,$lon) = @{$waypoints->{$k}};
	  $ret{$k} = [$lon,-$lat];
	}
	# determine boundaries of GPS coordinate rectangle
	my ($minlat,$maxlat,$minlon,$maxlon) = (100000,-100000,100000,-100000);
	foreach my $p (values %ret){
	  my ($lat,$lon) = @$p;
	  $minlat = ($lat<$minlat)?$lat:$minlat;
	  $minlon = ($lon<$minlon)?$lon:$minlon;
	  $maxlat = ($lat>$maxlat)?$lat:$maxlat;
	  $maxlon = ($lon>$maxlon)?$lon:$maxlon;
	}
	print "Latitudes  in ($minlat, $maxlat)\n";
	print "Longitudes in ($minlon, $maxlon)\n";

	my $target_width = $large_width;
	my $width  = $target_width/$window_zoom;
	my $height = (($maxlon-$minlon)*($target_width)/($maxlat-$minlat))/$window_zoom;
	my $offX   = -$window_offx/$window_zoom;
	my $offY   = -$window_offy/$window_zoom;
	#my $offX   = 746/$window_zoom-$window_offx/$window_zoom;
	#my $offY   = 191/$window_zoom-$window_offy/$window_zoom;
	$height *= 1.0000;
	$width  *= 1.0000;
	print "Scaling Waypoints to $width x $height +$offX+$offY\n";
	foreach my $k (keys %ret){
		my ($lat,$lon) = @{$ret{$k}};
		$lat = ($lat-$minlat)/($maxlat-$minlat)*($width) + $offX;
		$lon = ($lon-$minlon)/($maxlon-$minlon)*($height)+ $offY;
		$ret{$k} = [$lat,$lon]; 
	}
	return %ret;
}

sub determineSurroundingSupportPoints{ # <<<2
	# creates additional points in its
	my $gpspoints = shift;
	my $imgpoints = shift;

	# delete the points added last time
	map {delete $gpspoints->{$_}; delete $imgpoints->{$_}} grep {/^tmp/}keys %$gpspoints;

	# find "corner"-points
	my $maxmax = reduce{
			 ($gpspoints->{$a}->[0]+$gpspoints->{$a}->[1])
			>($gpspoints->{$b}->[0]+$gpspoints->{$b}->[1]) ? $a : $b
	}keys %$gpspoints;
	my $minmin = reduce{
			 ($gpspoints->{$a}->[0]+$gpspoints->{$a}->[1])
			<($gpspoints->{$b}->[0]+$gpspoints->{$b}->[1]) ? $a : $b
	}keys %$gpspoints;
	my $minmax = reduce{
			 (-$gpspoints->{$a}->[0]+$gpspoints->{$a}->[1])
			>(-$gpspoints->{$b}->[0]+$gpspoints->{$b}->[1]) ? $a : $b
	}keys %$gpspoints;
	my $maxmin = reduce{
			 ($gpspoints->{$a}->[0]-$gpspoints->{$a}->[1])
			>($gpspoints->{$b}->[0]-$gpspoints->{$b}->[1]) ? $a : $b
	}keys %$gpspoints;

	# create new points
	my @gpsadd = BaryCentric::getAddPoints( @$gpspoints{$maxmax,$minmin,$maxmin,$minmax});
	my @imgadd = BaryCentric::getAddPoints( @$imgpoints{$maxmax,$minmin,$maxmin,$minmax});

	# add them to the arguments
	foreach my $p (0..$#gpsadd){
		$gpspoints->{"tmp$p"} = $gpsadd[$p];
		$imgpoints->{"tmp$p"} = $imgadd[$p];
	}
}

1;
