#!/usr/bin/perl
# vim:fdm=marker:fmr=<<<,>>>

package RNDF; # <<<1
use strict;
use warnings;
use Carp;
use Data::Dumper;
use Smart::Comments;

use base qw(Class::ParseText::Base);
use Parse::RecDescent;

#$::RD_AUTOACTION = q{use Data::Dumper; print STDERR Dumper(@item)};
my $grammar = q(    # <<<2
  <autotree>
  comment   :      /\/\*[^\n]+/
  					{$item[1] =~ m|/\*(.*?)\*/|; $1}
  comment2  :       /\/\*[^\n]+/
  					{$item[1] =~ m|/\*(.*?)\*/|; $1}
  sep       :      /\s*[\n]+/
  string    :      /[\S]+/
  					{$item[1]}
  integer   :      /\d+/
  					{$item[1]}
  num_zones :      /\d+/
  					{$item[1]}
  num_lanes :      /\d+/
  					{$item[1]}
  num_segments  :  /\d+/
  					{$item[1]}
  num_waypoints :  /\d+/
  					{$item[1]}
  num_spots :  /\d+/
  					{$item[1]}
  id        :  /[\d.]+/
  					{$item[1]}
  latitude  :  /-?\d+\.\d+/
  					{$item[1]}
  longitude :  /-?\d+\.\d+/
  					{$item[1]}
  xcoor  :  /-?\d+\.?\d*/
  					{$item[1]}
  ycoor :  /-?\d+\.?\d*/
  					{$item[1]}

  waypoint  : id latitude longitude comment(?)
  exit_from : id  comment(?)
  exit_to   : id  comment(?)

  bline     : "double_yellow" | "solid_yellow" | "solid_white" | "broken_white"
  			{$item[1]}

  lane      : "lane" id comment(?) "num_waypoints" num_waypoints lane_opt(s?) waypoint(s) "end_lane"
  lane_opt  : "lane_width"     integer           comment(?)
            | "left_boundary"  bline             comment(?)
			| "right_boundary" bline             comment(?)
	    	| "checkpoint"     id integer        comment(?)
	    	| "stop"           id                comment(?)
	    	| "exit"           exit_from exit_to comment(?)
			| "zone_lane" comment(?)

  segment   : "segment" id comment(?) "num_lanes" num_lanes comment2(?) seg_opt(s?) lane(s) "end_segment"
  seg_opt   : "segment_name" string

  spot      : "spot" id spot_opt(s?) waypoint(s) "end_spot"
  spot_opt  : "spot_width" integer | "checkpoint" id integer

  perimeter : "perimeter" id comment(?) "num_perimeterpoints" integer peri_opt(s?) waypoint(s) "end_perimeter"
  peri_opt  : "exit" exit_from exit_to comment(?)

  zone      : "zone" id comment(?) "num_spots" num_spots zone_opt(s?) perimeter spot(s?) "end_zone"
  zone_opt  : "zone_name" string 

  filehead  : "RNDF_name" string "num_segments" num_segments "num_zones" num_zones filehead_opt(s?)
  filehead_opt : "format_version" string | "creation_date" string 
  extsupportpoint : string latitude longitude xcoor ycoor

startrule :  filehead segment(s?) zone(s?) "end_file" extsupportpoint(s) "end_extsupportpoint"
		  |  filehead segment(s?) zone(s?) "end_file" 
);

use RuleBase;                           # <<<2
@RNDF::startrule::ISA     = qw(RuleBase);
@RNDF::filehead::ISA      = qw(RuleBase);
@RNDF::filehead_opt::ISA  = qw(RuleBase);
@RNDF::zone::ISA          = qw(RuleBase);
@RNDF::zone_opt::ISA      = qw(RuleBase);
@RNDF::peri_opt::ISA      = qw(RuleBase);
@RNDF::peripoint::ISA     = qw(RuleBase);
@RNDF::perimeter::ISA     = qw(RuleBase);
@RNDF::spot::ISA          = qw(RuleBase);
@RNDF::spot_opt::ISA      = qw(RuleBase);
@RNDF::segment::ISA       = qw(RuleBase);
@RNDF::seg_opt::ISA       = qw(RuleBase);
@RNDF::lane::ISA          = qw(RuleBase);
@RNDF::lane_opt::ISA      = qw(RuleBase);
@RNDF::waypoint::ISA      = qw(RuleBase);
@RNDF::exit_from::ISA     = qw(RuleBase);
@RNDF::exit_to::ISA       = qw(RuleBase);
@RNDF::extsupportpoint::ISA  = qw(RuleBase);


sub init { # <<<2
	my $self = shift;
	$self->{parser}     = new Parse::RecDescent($grammar);
	$self->{start_rule} = "startrule";
	$self->{ensure_newline} = 0;
	return $self;
}

package RNDF::startrule; # <<<1
use Carp;
use Data::Dumper;
use Data::TreeDumper;
use Math::NumberCruncher;
use Smart::Comments '###';
use Math::Vec qw(:terse);
use List::Util  qw/min max reduce/;
use Data::Walk;

my @cntexits; # global var for walk (yuck!)

sub existsExitTo{ # <<<2
	my $self = shift;
	my $to   = shift;
	@cntexits = ();
	if ($to =~ /^\d+\.\d+\.\d+$/){
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_to"   and $_->id   eq $to)}, $self;
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_from" and $_->id   eq $to)}, $self;
	}elsif($to =~ /^\d+\.\d+$/){
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_to"   and $_->id12 eq $to)}, $self;
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_from" and $_->id12 eq $to)}, $self;
	}elsif($to =~ /^\d+$/){
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_to"   and $_->id1  eq $to)}, $self;
		walk sub{push @cntexits,$_->id if(ref $_ eq "RNDF::exit_from" and $_->id1  eq $to)}, $self;
	}else{
		die "Don't know how something can exit to '$to'";
	}
	return @cntexits if(scalar @cntexits>0);
	return ();
}

sub delCheckPoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	my $lane = $self->getLaneByWaypoint($wp);
	$lane->{'lane_opt(s?)'} = [ grep{($_->rule ne "checkpoint") or ($_->id ne $wp) or (!$_->inserted)}@{$lane->lane_opt} ];
}
sub addCheckPoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	my $checkpoints = shift;
	my $lane = $self->getLaneByWaypoint($wp);
	my $cpid    = max(values %$checkpoints) + 1;
	my @a = grep{$_->rule eq "checkpoint" and $_->id eq $wp}@{$lane->lane_opt};
	if (scalar @a){
		warn "This is a checkpoint already, won't overwrite.";
		return;
	}
	push @{$lane->lane_opt}, new RNDF::lane_opt( __STRING1__ => "checkpoint", id => $wp, integer => $cpid, comment => "Inserted");
}

sub delWayPoint{ # <<<2
	### delete waypoint
	my $self     = shift;
	my $wp       = shift;
	my $force    = shift;
	my %rename;

	my @e = $self->existsExitTo($wp);
	if (scalar @e >0){
		warn "Cannot delete Waypoint with connected exit @e!";
		return;
	}

	my $wpcontainer;
	if($self->isZone($wp))
	{
		$wpcontainer = $self->getZoneByWaypoint($wp)->perimeter;
		warn "Cannot delete points on zone perimeters";
		return;
	}else
	{
		$wpcontainer = $self->getLaneByWaypoint($wp);
		if(scalar @{$wpcontainer->waypoint} <= 2){
			### only one waypoint left --> delete lane
			return $self->delLaneByWaypoint($wp,$force);
		}
	}


	my ($wpnr)   = $wp =~ /^\d+\.\d+\.(\d+)$/;
	my $delwp    = $wpcontainer->waypoint->[$wpnr-1];
	unless ($delwp->inserted or $force){
		carp "Will not delete waypoint $wp, since it belongs to the original RNDF-file!";
		return {};
	}
	$self->delCheckPoint($wp);

	# create new array leaving out $wp
	my @newwp    = @{$wpcontainer->waypoint}[0..($wpnr-2)];
	push @newwp, map{
			$rename{$_->id} = $_->dec_nr; $_ # remember what to rename
		}@{$wpcontainer->waypoint}[($wpnr)..$#{$wpcontainer->waypoint}];
	$wpcontainer->{'waypoint(s)'} = \@newwp;
	$self->rename(\%rename);
	$wpcontainer->applyChange;
	return \%rename;
}
sub delLaneByWaypoint{ # <<<2
	### delete lane
	my $self = shift;
	my $wp   = shift;
	my $force = shift;
	my $seg   = $self->getSegmentByWaypoint($wp);
	my $lane  = $self->getLaneByWaypoint($wp);
	my $id2   = $lane->id2;
	my @e = $self->existsExitTo($lane->id);
	if (scalar @e > 0){
		warn "Cannot delete Lane with connected exit: @e!";
		return;
	}
	if (scalar @{$seg->lane} <= 1){
		$self->delSegmentByWaypoint($wp, $force);
		return;
	}
	$seg->{'lane(s)'} = [ grep{$lane!=$_}@{$seg->lane} ];
	my %rename = map{$_->id => ($_->id1 .".". ($_->id2 - 1))}grep{$_->id2 > $id2}@{$seg->lane};
	$self->rename(\%rename);
	return \%rename;
}
sub addSegment{ # <<<2
	### adding segment
	my $self                 = shift;
	my ($gx1,$gy1,$gx2,$gy2) = @_; # 1st waypoint

	my $p1   = V($gx1,$gy1);
	my $p2   = V($gx2,$gy2);
	my $p1p2 = $p2-$p1;
	my $dir  = V($$p1p2[1],-$$p1p2[0]);
	$dir     = $dir/$dir->Magnitude();
	$dir     = $dir/1E6;
	$dir     = $dir*20;

	my @lane2_1 = @{$dir+$p2}; # opposite direction
	my @lane2_2 = @{$dir+$p1};
	
	# increase all zone-ids
	my %ren     = map{$_->id => $_->id + 1}@{$self->zone};
	$self->rename(\%ren);

	my $segid   = max(map{$_->id}@{$self->segment}) + 1;
	print "inserting segment number $segid\n";

	my $wp11    = new RNDF::waypoint(id => $segid.".1.1", latitude => $gx1, longitude => $gy1, comment => 'Inserted');
	my $wp12    = new RNDF::waypoint(id => $segid.".1.2", latitude => $gx2, longitude => $gy2, comment => 'Inserted');

	my $wp21    = new RNDF::waypoint(id => $segid.".2.1", latitude => $lane2_1[0], longitude => $lane2_1[1], comment => 'Inserted');
	my $wp22    = new RNDF::waypoint(id => $segid.".2.2", latitude => $lane2_2[0], longitude => $lane2_2[1], comment => 'Inserted');

	my $lane1  = new RNDF::lane(    id => "$segid.1", comment => 'Inserted', 'waypoint(s)' => [$wp11,$wp12],   'lane_opt(s?)' => []);
	my $lane2  = new RNDF::lane(    id => "$segid.2", comment => 'Inserted', 'waypoint(s)' => [$wp21,$wp22],   'lane_opt(s?)' => []);
	my $seg    = new RNDF::segment( id =>  $segid,    comment => 'Inserted', 'lane(s)'     => [$lane1,$lane2], 'seg_opt(s?)'  => []);

	push @{$self->segment}, $seg;
}
sub addZone{ # <<<2
	### adding segment
	my $self                 = shift;
	my ($gx1,$gy1,$gx2,$gy2) = @_; # 1st waypoint

	my $p1   = V($gx1,$gy1);
	my $p2   = V($gx2,$gy2);

	my $zid     = max(map{$_->id}@{$self->zone}) + 1;
	### inserting zone number: $zid
	my $peri    = new RNDF::perimeter(id=>"$zid.0", 'waypoint(s)' => [],  'peri_opt(s?)' => [], comment => "Inserted");
	my $radius  = ($p2-$p1)->Magnitude();
	my $num     = 20;
	foreach my $n (1..$num){
		my $x = $$p1[0] + $radius * cos($n*2*3.14/$num);
		my $y = $$p1[1] + $radius * sin($n*2*3.14/$num);
		push @{$peri->waypoint}, new RNDF::waypoint(id=>"$zid.0.$n", latitude => $x, longitude => $y, 'comment' => "Inserted");
	}

	my $zone = new RNDF::zone(id => $zid, 'perimeter' => $peri, 'spot(s)' => [], 'zone_opt(s?)' => [], comment => "Inserted");

	push @{$self->zone}, $zone;
}
sub addSpot{ # <<<2
	my $self = shift;
	my $gx1  = shift;
	my $gy1  = shift;
	my $gx2  = shift;
	my $gy2  = shift;
	my $checkpoints = shift;
	my $zone = undef;
	foreach my $z (@{$self->zone}){
		my @perim = map {($_->latitude,$_->longitude)} @{$z->perimeter->waypoint};
		if (Math::NumberCruncher::InPolygon($gx1,$gy1,@perim)){
			$zone = $z;
			### Spot must be in zone: $zone->id
			last;
		}
	}
	unless($zone){
		warn "Cannot find the zone you clicked in.";
		return;
	}
	my $cpid    = max(values %$checkpoints) + 1;
	my $zid     = $zone->id;
	my $nums    = max(map{$_->id2}@{$zone->spot})+1;
	my $wp11    = new RNDF::waypoint(id => "$zid.$nums.1", latitude => $gx1, longitude => $gy1, comment => 'Inserted');
	my $wp12    = new RNDF::waypoint(id => "$zid.$nums.2", latitude => $gx2, longitude => $gy2, comment => 'Inserted');
	my $opt     = new RNDF::spot_opt('__STRING1__' => "checkpoint", id => "$zid.$nums.1", integer => $cpid, comment => "Inserted"); 
	push @{$zone->spot}, new RNDF::spot(id => "$zid.$nums", 'spot_opt(s?)' => [$opt], 'waypoint(s)' => [$wp11,$wp12]);
}
sub getZoneByWaypoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	#### retrieving zone for waypoint: $wp
	my ($s,$l,$p) = $wp =~ m/^(\d+).(\d+).(\d+)/;
	my $maxseg    = max(map{$_->id}@{$self->segment});
	my $zone      = $self->zone->[$s-$maxseg-1];
	my $id = $zone->id;
	die "Zone ID $id does not match requested one ($s)" unless ($id eq "$s");
	return $zone;
}
sub getSegmentByWaypoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	#### retrieving segment for waypoint: $wp
	my ($s,$l,$p) = $wp =~ m/^(\d+).(\d+).(\d+)/;
	my $seg = $self->segment->[$s-1];
	my $id = $seg->id;
	die "Segment ID $id does not match requested one ($s)" unless ($id eq "$s");
	return $seg;
}
sub delZoneByWaypoint{ # <<<2
	### deleting zone
	my $self   = shift;
	my $wp     = shift;
	my $force  = shift;
	my $zone   = $self->getZoneByWaypoint($wp);
	if ($self->existsExitTo($zone->id)){
		warn "Cannot delete Segment with connected exit!";
		return;
	}
	unless ($zone->inserted or $force){
		print "Cannot delete a zone from original RNDF file!\n";
		return {};
	}
	my $id     = $zone->id;
	$self->{zone} = [ grep{$zone!=$_}@{$self->zone} ];
	my %rename = map{$_->id => $_->id - 1}grep{$_->id > $id}@{$self->zone};
	$self->rename(\%rename);
	return \%rename;
}
sub delSegmentByWaypoint{ # <<<2
	### deleting segment 
	my $self   = shift;
	my $wp     = shift;
	my $force  = shift;
	my $seg    = $self->getSegmentByWaypoint($wp);
	if ($self->existsExitTo($seg->id)){
		warn "Cannot delete Segment with connected exit!";
		return;
	}
	unless ($seg->inserted or $force){
		print "Cannot delete segment of $wp from original RNDF file!\n";
		return {};
	}
	my $id     = $seg->id;
	$self->{segment} = [ grep{$seg!=$_}@{$self->segment} ];
	my %rename = map{$_->id => $_->id - 1}grep{$_->id > $id}@{$self->segment},@{$self->zone};
	$self->rename(\%rename);
	return \%rename;
}
sub getWaypointByWaypoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	### retrieving waypoint for waypoint: $wp
	confess "Not a waypoint " unless ($wp =~ m/^(\d+).(\d+).(\d+)/);
	my ($s,$l,$p) = $wp =~ m/^(\d+).(\d+).(\d+)/;
	$wp = $self->segment->[$s-1]->lane->[$l-1]->waypoint->[$p-1];
	my $id = $wp->id;
	die "Waypoint ID $id does not match requested one ($s.$l.$p)" unless ($id eq "$s.$l.$p");
	return $wp;
}
sub isZone{ # <<<2
	my $self     = shift;
	my $id       = shift;
	$id =~ s/\..*$//g;
	my $maxseg   = max(map{$_->id}@{$self->segment});
	if ($id > $maxseg){
		return 1;
	}
	return 0;
}
sub getLaneByWaypoint{ # <<<2
	my $self = shift;
	my $wp   = shift;
	### retrieving lane for waypoint: $wp
	confess "Not a waypoint " unless ($wp =~ m/^(\d+).(\d+).(\d+)/);
	my ($s,$l,$p) = $wp =~ m/^(\d+).(\d+).(\d+)/;
	my $lane = $self->segment->[$s-1]->lane->[$l-1];
	my $id = $lane->id;
	die "Lane ID $id does not match requested one ($s.$l)" unless ($id eq "$s.$l");
	return $lane;
}
sub moveWaypoint{ # <<<2
	my $self         = shift;
	my ($wp,$gx,$gy,$force) = @_;
	### moving waypoint in tree: $wp
	map{$_->moveWaypoint($wp,$gx,$gy, $force)}@{$self->segment};
	map{$_->moveWaypoint($wp,$gx,$gy, $force)}@{$self->zone};
}
sub rndf{ # <<<2
	my $self = shift;
	my $s = "RNDF_name\t"      . $self->filehead->string        . "\n";
	$s .=   "num_segments\t"   . scalar(@{$self->segment})      . "\n";
	$s .=   "num_zones\t"      . scalar(@{$self->zone})         . "\n";
	foreach (@{$self->filehead->filehead_opt}){
		$s .=   "format_version\t" . $_->string . "\n" if ($_->rule eq "format_version");
		$s .=   "creation_date\t"  . $_->string . "\n" if ($_->rule eq "creation_date");
	}
	$s .= "\n";
	$s .= join("\n\n\n", map{$_->rndf}@{$self->segment});
	$s .= "\n";
	$s .= "\n";
	$s .= join("\n\n", map{$_->rndf}@{$self->zone});
	$s .= "\nend_file\n";
	if(ref($self->{'extsupportpoint(s)'})){
		$s .= "/* Move the following block to the end of anther RNDF file to align it in the same picture */\n";
		foreach my $esp (@{$self->extsupportpoint}){
			$s .= "/* EXTENDEDSUPPORTPOINT " . join("\t",$esp->string , $esp->latitude , $esp->longitude, $esp->xcoor, $esp->ycoor) . " */\n";
		}
		$s .= "/* EXTENDEDSUPPORTPOINT end_extsupportpoint */\n";
	}
	$s;
}
my %renamedIDs;
sub rename { # <<<2
	### Renaming stuff in tree
	my $self     = shift;
	my $rename   = shift;
	%renamedIDs  = ();
	print DumpTree($rename);
	#$self->visit(\&renameFunction,$rename);
	walk sub{renameFunction($_, $rename); 1}, $self;
}
sub renameFunction { # <<<2
	# this can rename an arbitrary object according to the hash %$ren
	# it will also rename parts, that is,
	#     if %ren = (1=>2) then 1.5.1 becomes 2.5.1
	my $obj = shift;
	my $ren = shift;
	return unless(ref($obj) =~ /^RNDF::/);
	return unless exists $obj->{id};
	my $id = \$obj->{id};
	if($$id !~ /^[\d.]+/){
		#print "ignoring id=$$id for renaming\n";
		return 0;
	}
	my $num = $$id =~ s/\././g;
	my $prev = $renamedIDs{$$id};
	if($prev and $prev eq ref($obj) and $prev =~/segment/){
		warn "FIXME: Previously renamed $prev $$id, ignoring rename request\n";
		return;
	}
	$renamedIDs{$$id} = ref $obj;
	if ($num == 0){
		#need to rename segment/zone?
		if ($$ren{$$id}){
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$$id};
		}else{
			return 0;
		}
	}elsif($num == 1){
		# need to rename lane/spot?
		if ($$ren{$$id}){
			# directly
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$$id} 
		}elsif($$ren{$obj->id1}){
			# via segment/zone
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$obj->id1} . "." . $obj->id2;
		}else{
			return 0;
		}
	}elsif($num == 2){
		# need to rename waypoint/perimeterwaypoint/spotwaypoint?
		if ($$ren{$$id}){
			# directly
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$$id} 
		}elsif($$ren{$obj->id1}){
			# via segment/zone
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$obj->id1} . "." . $obj->id23;
		}elsif($$ren{$obj->id12}){
			# via lane/spot
			#print "renaming ",ref($obj)," $$id\n";
			$$id = $$ren{$obj->id12} . "." . $obj->id3;
		}else
		{
			return 0;
		}
	}
	#print "   object id is now $$id\n";
	return 1;
}

sub removeWaypointsWithFlag{ # <<<2
	my $self = shift;
	my $flag = shift;
	walk sub{ref($_) =~ /^RNDF::waypoint/ and $_->{comment} and $_->comment =~ /$flag/i and $self->delWayPoint($_->id)}, $self;
}
package RNDF::segment; # <<<1
use Data::Dumper;
use Smart::Comments;
sub moveWaypoint{ # <<<2
	my $self = shift;
	my ($wp,$gx,$gy,$force) = @_;
	my ($sid,$lid,$wpid) = $wp =~ /^(\d+)\.(\d+)\.(\d+)$/;
	return unless($self->id eq $sid);
	### moving waypoint in segment: $wp
	foreach my $l (@{$self->lane}){
		my %stoppoints;
		my %exitpoints;
		my %checkpoints;
		foreach my $o (@{$l->lane_opt}){
			$stoppoints{$o->id}=1            if ($o->rule eq "stop");
			$exitpoints{$o->exit_from->id}=1 if ($o->rule eq "exit");
			$exitpoints{$o->exit_to->id}=1   if ($o->rule eq "exit");
			$checkpoints{$o->id}=1           if ($o->rule eq "checkpoint");
		}
		next unless($l->id2 eq $lid);
		$wp = $l->waypoint->[$wpid-1];
		if(!$force and !$wp->inserted and $wp->id3 != 1 and (!$exitpoints{$wp->id} or $stoppoints{$wp->id} or $checkpoints{$wp->id})){
			warn "Cannot move points from original RNDF file!"
			    ."UNLESS they are exits and neither supportpoints nor stops";
			return;
		}
		my $ox= $wp->{latitude} ;
		my $oy= $wp->{longitude};
		if(!$wp->inserted and exists $wp->{comment} and $wp->{comment} !~ /Moved/i){
			# inserting comment to mark moved waypoint...
			$wp->{comment}   = "Moved($ox;$oy)"
		}
		$wp->{latitude}  = $gx;
		$wp->{longitude} = $gy;
		return;
	}
}
sub rndf{ # <<<2
	my $self = shift;
	my $s = "segment\t"   . $self->id   ."\t".$self->comment. "\n";
	$s .=   "num_lanes\t" . scalar(@{$self->lane}) . $self->comment2 . "\n";
	foreach (@{$self->seg_opt}){
		$s .=   "segment_name\t" . $_->string . "\n" if ($_->rule eq "segment_name");
	}
	$s .= join("\n\n", map{$_->rndf}@{$self->lane});
	$s .= "\nend_segment";
	$s;
}
package RNDF::lane; # <<<1
use Carp;
use Math::Vec qw(:terse);
use List::Util  qw/min max reduce/;
use Data::Dumper;
use Smart::Comments;

sub applyChange{ # <<<2
	my $self     = shift;
	map{$_->applyChange} @{$self->waypoint};
}
sub addWayPoint{ # <<<2
	my $self     = shift;
	my $wp       = V($_[0],$_[1]);
	my @lines;
	foreach my $p (0..($#{$self->waypoint}-1)){
		push @lines,[${$self->waypoint}[$p],${$self->waypoint}[$p+1]];
	}
	my $line = undef;
	my $proj = undef;
	my $dist = 1E6;
	foreach my $l(@lines){
		my $A = V($$l[0]->coor);
		my $B = V($$l[1]->coor);
		my $C = V($wp->PerpFoot($A,$B));
		if (V($C->Minus($A))->Magnitude()+V($C->Minus($B))->Magnitude() < V($A->Minus($B))->Magnitude()+0.0000001){
			my $d = (V($C->Minus($wp))->Magnitude());
			if ($dist> $d){
				$line = $l;
				$proj = $C;
				$dist = $d;
			}
		}
	}
	### adding waypoint between: join('; ', $$line[0]->id , $$line[1]->id ) 
	my @nswp    = sort{$a->id3 <=> $b->id3 } @{$self->waypoint};

	my $nwp     = new RNDF::waypoint(
			'id'         => "_" . $self->id . "." . $$line[1]->id3,
			'latitude'   => $$proj[0],
			'longitude'  => $$proj[1],
			'comment'    => "Inserted");
	### as rndf: $nwp->rndf

	# construct new array of points. First include ones _before_
	my @newwp   = @nswp[0..($$line[0]->id3-1)];
	# ... the new point
	push @newwp, $nwp;
	# and the following ones
	my %renamed;
	push @newwp, map{
			$renamed{$_->id}=$_->inc_nr; # request renaming
			$_;
		}@nswp[($$line[1]->id3-1) .. $#nswp];
	$self->{'waypoint(s)'} = \@newwp;
	return \%renamed;
}
sub rndf { # <<<2
	my $self = shift;
	my $s = "lane\t"          . $self->id   . "\t".$self->comment . "\n";
	$s .=   "num_waypoints\t" . scalar(@{$self->waypoint})  . "\n";
	foreach (@{$self->lane_opt}){
		$s .=   "zone_lane\t"                           .$_->comment    ."\n"                   if ($_->rule eq "zone_lane");
		$s .=   "lane_width\t"     . $_->integer   ."\t".$_->comment    ."\n"                   if ($_->rule eq "lane_width");
		$s .=   "left_boundary\t"  . $_->bline     ."\t".$_->comment    ."\n"                   if ($_->rule eq "left_boundary");
		$s .=   "right_boundary\t" . $_->bline     ."\t".$_->comment    ."\n"                   if ($_->rule eq "right_boundary");
		$s .=   "stop\t"       . $_->id            ."\t".$_->comment    ."\n"                   if ($_->rule eq "stop");
		$s .=   "checkpoint\t" . $_->id            ."\t".$_->integer    ."\t".$_->comment ."\n" if ($_->rule eq "checkpoint");
		$s .=   "exit\t"       . $_->exit_from->id ."\t".$_->exit_to->id."\t".$_->comment ."\n" if ($_->rule eq "exit");
	}
	$s .= join("\n", map{$_->rndf}@{$self->waypoint});
	$s .= "\nend_lane";
	$s;
}
package RNDF::waypoint; # <<<1
use Carp;
use Data::Dumper;
sub rndf{ # <<<2
	my $self = shift;
	my $s    = sprintf("%s\t%3.6f\t%3.6f" ,$self->id, $self->latitude , $self->longitude);
	$s      .= "\t".$self->comment;
	return $s;
}
sub dec_nr{ # <<<2
	my $self = shift;
	$self->{id} =~ /^(\d+)\.(\d+)\.(\d+)$/ or confess "Waypoint rule $self->{id} does not match pattern";
	return "$1.$2.".($3-1);
}
sub inc_nr{ # <<<2
	my $self = shift;
	$self->{id} =~ /^(\d+)\.(\d+)\.(\d+)$/ or confess "Waypoint rule $self->{id} does not match pattern:" . Dumper($self->{id});
	return "$1.$2.".($3+1);
}
sub coor{ # <<<2
	return ($_[0]->latitude, $_[0]->longitude)
}
sub dist{  # <<<2 # <<<2
	my $self = shift;
	return sqrt($self->dist2(@_));
}
sub dist2{ # <<<2 # <<<2
	my $self = shift;
	my ($a,$b) = @_;
	return ($self->latitude-$a)**2 + ($self->longitude-$b)**2
}
sub applyChange{ # <<<2
	my $self = shift;
	$self->{id} =~ /^_(.+)$/ or return;
	$self->{id} = $1;
	return $self;
}

sub supportpoint{ # <<<2
	my $self = shift;
	if (scalar @_ and defined $_[0]){
		# Setting supportpoint: $self->id
		$self->{comment} = "SupportPoint $_[0] $_[1]";
		return 1;
	}elsif(scalar @_){
		# Clearing supportpoint: $self->id
		$self->{comment} = "";
		return 1;
	}
	return unless (exists $self->{comment});
	return unless ($self->{comment} =~ /^SupportPoint ([\d.]+) ([\d.]+)$/);
	# returning supportpoint: $self->id
	# returning supportpoint: $self->{comment}
	return ($1,$2);
}
package RNDF::zone; # <<<1
sub moveWaypoint{ # <<<2
	my $self = shift;
	my ($wp,$gx,$gy,$force) = @_;
	$wp =~ /^(\d+)\.(\d+)\.(\d+)$/;
	return unless($self->id eq "$1");
	if ($2 == 0){
		$wp = $self->perimeter->waypoint->[$3-1];
	}else{
		$wp = $self->spot->[$2-1]->waypoint->[$3-1];
	}
	if (!$force and !$wp->inserted){
		warn "cannot move non-inserted waypoint";
		return;
	}
	$wp->{latitude}  = $gx;
	$wp->{longitude} = $gy;
}
sub rndf{ # <<<2
	my $self = shift;
	my $s = "zone\t"      . $self->id    . "\t" . $self->comment. "\n";
	$s .=   "num_spots\t" . scalar(@{$self->spot}) . "\n";
	foreach (@{$self->zone_opt}){
		$s .=   "zone_name\t" . $_->string . "\n" if ($_->rule eq "zone_name");
	}
	$s .= $self->perimeter->rndf($self->id . ".0") . "\n";
	my $i = 1;
	$s .= join("\n", map{$_->rndf($self->id . ".".($i++))}@{$self->spot});
	$s .= "\nend_zone";
	$s;
}
package RNDF::perimeter; # <<<1
sub rndf{ # <<<2
	my $self = shift;
	my $id   = shift;
	my $s = "perimeter\t" . $id . "\t".$self->comment ."\n";
	$s .= "num_perimeterpoints\t" . scalar(@{$self->waypoint}) . "\n";
	foreach (@{$self->peri_opt}){
		$s .=   "exit\t" . $_->exit_from->id ."\t". $_->exit_to->id . "\t" . $_->comment. "\n" if ($_->rule eq "exit");
	}
	$s .= join("\n", map{$_->rndf}@{$self->waypoint});
	$s .= "\nend_perimeter";
	$s;
}

package RNDF::spot; # <<<1
sub rndf{ # <<<2
	my $self = shift;
	my $id   = shift; # id
	my $s = "spot\t" . $id ."\t".$self->comment. "\n";
	foreach (@{$self->spot_opt}){
		$s .=   "spot_width\t" . $_->integer                      . "\n" if ($_->rule eq "spot_width");
		$s .=   "checkpoint\t" . $_->id      . "\t" . $_->integer . "\n" if ($_->rule eq "checkpoint");
	}
	$s .= join("\n", map{$_->rndf}@{$self->waypoint});
	$s .= "\nend_spot";
	$s;
}

package RndfParser; # <<<1

use RNDF;
use Parse::RecDescent::Topiary;
use Data::Dumper;
use Data::TreeDumper;

sub get{ # <<<2
	my $rndffile = shift;

	### loading rndf file...
	open R, "<$rndffile" or die "could not open $rndffile";
	my $t;
	$t .= $_ while(<R>);
	close R;

	#### parsing rndf file...
	$t =~ s|^\s*/\*\s*EXTENDEDSUPPORTPOINT\s*(.*?)\*/\s*$|\1|mg;
	$t =~ s|^\s*/\*.*?\*/\s*$||mg;
	my $p = RNDF->new;
	my $r = $p->parse_text($t);
	die "Syntax Error in '$rndffile'" unless($r);
	my $tree = topiary(
		tree        => $r,
		namespace   => 'RNDF',
		consolidate => 1
	);
	die "could not parse file!" unless($tree);
	### dumping parsed tree...
	#open F, ">t/sitevisit.dump" || die $!;
	#print F DumpTree($tree);
	#close F;
	$tree->{startrule}
}

1;
